package com.alterdekim.game.service;

import com.alterdekim.game.component.game.*;
import com.alterdekim.game.component.game.inventory.Inventory;
import com.alterdekim.game.component.game.inventory.InventoryList;
import com.alterdekim.game.component.game.inventory.InventoryTab;
import com.alterdekim.game.entity.Role;
import com.alterdekim.game.entity.User;
import com.alterdekim.game.entity.UserProperty;
import com.alterdekim.game.repository.RoleRepository;
import com.alterdekim.game.repository.UserPropertyRepository;
import com.alterdekim.game.repository.UserRepository;
import lombok.AllArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

@Service
public class UserService {

    @Autowired
    private UserRepository userRepository;
    @Autowired
    private RoleRepository roleRepository;
    @Autowired
    private PasswordEncoder passwordEncoder;
    @Autowired
    private UserPropertyRepository propertyRepository;
    @Autowired
    private SmileService smileService;

    @Autowired
    private AvatarInventoryService inventoryService;

    public User findByUsername(String username) {
        return userRepository.findByUsername(username);
    }

    public long saveUser(String username, String password) {
        User user = new User();
        user.setUsername(username);
        user.setPassword(passwordEncoder.encode(password));
        Role role = roleRepository.findByName("ROLE_USER");
        if(role == null){
            role = checkRoleExist();
        }
        user.setRoles(Collections.singletonList(role));
        return userRepository.save(user).getId();
    }

    private Role checkRoleExist() {
        Role role = new Role();
        role.setName("ROLE_USER");
        return roleRepository.save(role);
    }

    public Optional<Object> findUserProperty(Long userId, PlayerProperties type) {
        Optional<UserProperty> res = this.propertyRepository.findByUserIdAndTag(userId, type);
        if( res.isEmpty() ) return Optional.empty();
        UserProperty property = res.get();
        return Optional.of(switch (property.getType().getValueType()) {
            case String -> property.getStringVal();
            case Integer -> property.getIntegerVal();
            case Boolean -> property.getBooleanVal();
            case Date -> property.getDateVal();
            case Enum -> property.getType() == PlayerProperties.RoleFlags ? RoleFlags.valueOf(property.getStringVal()) : ClubAccessType.valueOf(property.getStringVal());
        });
    }

    public void pushUserProperty(Long userId, PlayerProperties type, Object val) {
        this.propertyRepository.deleteByUserIdAndType(userId, type);
        this.propertyRepository.save(switch (type.getValueType()) {
            case String -> new UserProperty(userId, (String) val, null, null, null, type);
            case Integer -> new UserProperty(userId, null, null, (Integer) val, null, type);
            case Boolean -> new UserProperty(userId, null, (Boolean) val, null, null, type);
            case Date -> new UserProperty(userId, null, null, null, (LocalDateTime) val, type);
            case Enum -> new UserProperty(userId, String.valueOf(val), null, null, null, type);
        });
    }

    // todo: optimize all getters, group them by type of result value.
    public RoleFlags getRoleFlags(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.RoleFlags);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.RoleFlags, RoleFlags.SA);
            return RoleFlags.SA;
        }
        return (RoleFlags) res.get();
    }

    public Boolean getIsLimited(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.IsLimited);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.IsLimited, false);
            return false;
        }
        return (Boolean) res.get();
    }

    public Boolean getIsMinorSecretAgent(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.IsMinorSecretAgent);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.IsMinorSecretAgent, true);
            return false;
        }
        return (Boolean) res.get();
    }

    public Boolean getIsHomeLocked(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.IsHomeLocked);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.IsHomeLocked, true);
            return true;
        }
        return (Boolean) res.get();
    }

    public Boolean getIsClubLocked(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.IsClubLocked);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.IsClubLocked, true);
            return true;
        }
        return (Boolean) res.get();
    }

    public Boolean getIsClubPresent(Long userId) {
        // todo: Implement clubs;
        return false;
    }

    public ClubAccessType getClubAccessType(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.ClubAccessType);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.ClubAccessType, ClubAccessType.CLOSED);
            return ClubAccessType.CLOSED;
        }
        return (ClubAccessType) res.get();
    }

    public Boolean getIsFriend(Long userId) {
        // todo: Implement friends;
        return false;
    }

    public Integer getExperience(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.Experience);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.Experience, 0);
            return 0;
        }
        return (Integer) res.get();
    }

    public Integer getLevel(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.Level);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.Level, 1);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getMagicLevel(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.MagicLevel);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.MagicLevel, 1);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getRaceLevel(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.RaceLevel);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.RaceLevel, 0);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getCampId(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.CampId);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.CampId, 0);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getNiftTotalCount(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.NiftTotalCount);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.NiftTotalCount, 0);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getInstructorLevel(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.InstructorLevel);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.InstructorLevel, 10);
            return 1;
        }
        return (Integer) res.get();
    }

    public Integer getUsualTickets(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.UsualTickets);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.UsualTickets, 1000);
            return 1000;
        }
        return (Integer) res.get();
    }

    public Integer getMagicTickets(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.MagicTickets);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.MagicTickets, 500);
            return 500;
        }
        return (Integer) res.get();
    }

    public Integer getVisaId(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.VisaId);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.VisaId, 10);
            return 0;
        }
        return (Integer) res.get();
    }

    public Integer getCurrentUseCount(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.CurrentUseCount);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.CurrentUseCount, 10);
            return 0;
        }
        return (Integer) res.get();
    }

    public Integer getLastNewspaperId(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.LastNewspaperId);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.LastNewspaperId, 10);
            return 0;
        }
        return (Integer) res.get();
    }

    public Boolean getHasSnFriends(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.HasSnFriends);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.HasSnFriends, true);
            return false;
        }
        return (Boolean) res.get();
    }

    public Integer getMinsLeft(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.MinsLeft);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.MinsLeft, 10);
            return 0;
        }
        return (Integer) res.get();
    }

    public Integer getTimesUsed(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.TimesUsed);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.TimesUsed, 10);
            return 0;
        }
        return (Integer) res.get();
    }

    public Integer getPhoneCardBalance(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.PhoneCardBalance);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.PhoneCardBalance, 10);
            return 10;
        }
        return (Integer) res.get();
    }

    public Integer getWeaponsCount(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.WeaponsCount);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.WeaponsCount, 15);
            return 15;
        }
        return (Integer) res.get();
    }

    public void setWeaponsCount(Long userId, Integer count) {
        this.pushUserProperty(userId, PlayerProperties.WeaponsCount, count);
    }

    public Integer getOxygenUnits(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.OxygenUnits);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.OxygenUnits, 1);
            return 1;
        }
        return (Integer) res.get();
    }

    public Optional<String> getUsernameById(Long userId) {
        return this.userRepository.findById(userId).map(User::getUsername);
    }

    public UserAccount getUserAccountByUserId(Long userId) {
        return new UserAccount(
                getPhoneCardBalance(userId),
                getWeaponsCount(userId),
                getOxygenUnits(userId)
        );
    }

    public UserPhone getUserPhoneByUserId(Long userId) {
        return new UserPhone(
                0,
                0L,
                0,
                getPhoneCardBalance(userId)
        );
    }

    public Inventory getUserAvatarByUserId(Long userId) {
        return new Inventory(
            new InventoryTab(199, List.of(
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Body),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Legs)
            )),
            new InventoryTab(205, List.of(
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Horn),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Beak),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Mouth),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Nose),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Eyes),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Ears)
            )),
            new InventoryTab(206, List.of(
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Clothes),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Caps),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Glasses),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Boots),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Bangles),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Suits)
            )),
            new InventoryList(230, Collections.emptyList()), // todo: implement that
            new InventoryList(231, Collections.emptyList()),
            new InventoryList(211, Collections.emptyList())
        );
    }

    public UserInfo getUserInfoByUserId(Long userId, boolean isForUserInit) {
        // todo: implement hardcoded stuff
        return new UserInfo(
                userId,
                !isForUserInit ? userId : null,
                getRoleFlags(userId),
                !isForUserInit ? true : null,
                getIsLimited(userId),
                true,
                getIsMinorSecretAgent(userId),
                getIsHomeLocked(userId),
                getIsClubLocked(userId),
                getIsClubPresent(userId),
                getClubAccessType(userId),
                null,
                getIsFriend(userId),
                getExperience(userId),
                getLevel(userId),
                getMagicLevel(userId),
                getRaceLevel(userId),
                5,
                10,
                getCampId(userId),
                getNiftTotalCount(userId),
                "Server",
                LocalDateTime.now(),
                getInstructorLevel(userId),
                getUsualTickets(userId),
                getMagicTickets(userId),
                getVisaId(userId),
                getCurrentUseCount(userId),
                getLastNewspaperId(userId),
                getHasSnFriends(userId),
                getMinsLeft(userId),
                getTimesUsed(userId),
                this.getUsernameById(userId).get(),
                "110,111,112",
                isForUserInit ? null : inventoryService.getMagicAbilitiesItemsByUserId(userId)
        );
    }

    public List<BodyAvatarItem> getOtherUserAvatar(Long userId) {
        return this.inventoryService.getWholeAvatarByUserId(userId);
    }

    public UserInitInfo getUserInitInfoByUserId(Long userId) {
        return new UserInitInfo(
                getUserAccountByUserId(userId),
                getUserInfoByUserId(userId, true),
                getUserPhoneByUserId(userId),
                inventoryService.getMagicAbilitiesItemsByUserId(userId),
                smileService.findAll()
        );
    }

    public PlayerAvatar getAvatarById(long userId, Player p) {
        Map<String, BodyAvatarItem> body = new HashMap<>();
        List<BodyAvatarItem> i = inventoryService.getAvatarItemsByUserId(userId);
        IntStream.range(0, i.size()).boxed().forEach(g -> body.put(String.valueOf(g+1), i.get(g)));
        UserAvatar a = new UserAvatar(getLevel(userId), getUsernameById(userId).get(), p.getX(), p.getY(), p.getState(), body);
        return new PlayerAvatar(String.valueOf(getRoleFlags(userId).getValue()), a);
    }

    public void removeUser(long userId) {
        this.userRepository.deleteById(userId);
    }
}
